@@ -8,6 +8,7 @@ import re
 import requests
 from enum import Enum, auto
 from typing import Any, Dict
+import os
 
 import cloudscraper
 
@@ -23,7 +24,7 @@ class ApiClient:
         "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:66.0) Gecko/20100101 Firefox/66.0"
     }
 
-    def __init__(self, session, baseurl, headers=None, aditional_headers=None):
+    def __init__(self, session, baseurl, headers=None, additional_headers=None):
         """Return a new Client instance."""
         self.session = session
         self.baseurl = baseurl
@@ -33,8 +34,8 @@ class ApiClient:
         else:
             self.headers = self.default_headers.copy()
 
-        if aditional_headers:
-            self.headers.update(aditional_headers)
+        if additional_headers:
+            self.headers.update(additional_headers)
 
     def set_cookies(self, cookies):
         logger.debug("Restoring cookies for saved session")
@@ -55,11 +56,11 @@ class ApiClient:
 
         return path
 
-    def get(self, addurl, aditional_headers=None, params=None):
+    def get(self, addurl, additional_headers=None, params=None):
         """Make an API call using the GET method."""
         total_headers = self.headers.copy()
-        if aditional_headers:
-            total_headers.update(aditional_headers)
+        if additional_headers:
+            total_headers.update(additional_headers)
         url = self.url(addurl)
 
         logger.debug("URL: %s", url)
@@ -81,11 +82,11 @@ class ApiClient:
 
             raise GarminConnectConnectionError(err) from err
 
-    def post(self, addurl, aditional_headers, params, data):
+    def post(self, addurl, additional_headers=None, params=None, data=None, files=None):
         """Make an API call using the POST method."""
         total_headers = self.headers.copy()
-        if aditional_headers:
-            total_headers.update(aditional_headers)
+        if additional_headers:
+            total_headers.update(additional_headers)
         url = self.url(addurl)
 
         logger.debug("URL: %s", url)
@@ -94,7 +95,7 @@ class ApiClient:
 
         try:
             response = self.session.post(
-                url, headers=total_headers, params=params, data=data
+                url, headers=total_headers, params=params, data=data, files=files
             )
             response.raise_for_status()
             # logger.debug("Response: %s", response.content)
@@ -171,8 +172,12 @@ class Garmin:
             "proxy/wellness-service/wellness/dailyStress"
         )
 
+        self.garmin_connect_goals_url = "proxy/goal-service/goal/goals"
+
         self.garmin_connect_rhr_url = "proxy/userstats-service/wellness/daily"
 
+        self.garmin_connect_hrv_url = "proxy/hrv-service/hrv"
+
         self.garmin_connect_training_readiness_url = "proxy/metrics-service/metrics/trainingreadiness"
 
         self.garmin_connect_training_status_url = "proxy/metrics-service/metrics/trainingstatus/aggregated"
@@ -193,6 +198,9 @@ class Garmin:
             "proxy/activitylist-service/activities/search/activities"
         )
         self.garmin_connect_activity = "proxy/activity-service/activity"
+        self.garmin_connect_activity_types = "proxy/activity-service/activity/activityTypes"
+
+        self.garmin_connect_fitnessstats = "proxy/fitnessstats-service/activity"
 
         self.garmin_connect_fit_download = "proxy/download-service/files/activity"
         self.garmin_connect_tcx_download = "proxy/download-service/export/tcx/activity"
@@ -200,9 +208,10 @@ class Garmin:
         self.garmin_connect_kml_download = "proxy/download-service/export/kml/activity"
         self.garmin_connect_csv_download = "proxy/download-service/export/csv/activity"
 
-        self.garmin_connect_fit_upload_url = "proxy/upload-service/upload/.fit"
+        self.garmin_connect_upload = "proxy/upload-service/upload"
 
         self.garmin_connect_gear = "proxy/gear-service/gear/filterGear"
+        self.garmin_connect_gear_baseurl = "proxy/gear-service/gear/"
 
         self.garmin_connect_logout = "auth/logout/?url="
 
@@ -212,12 +221,12 @@ class Garmin:
         self.sso_rest_client = ApiClient(
             self.session,
             self.garmin_connect_sso_url,
-            aditional_headers=self.garmin_headers,
+            additional_headers=self.garmin_headers,
         )
         self.modern_rest_client = ApiClient(
             self.session,
             self.garmin_connect_modern_url,
-            aditional_headers=self.garmin_headers,
+            additional_headers=self.garmin_headers,
         )
 
         self.display_name = None
@@ -576,6 +585,14 @@ class Garmin:
 
         return self.modern_rest_client.get(url, params=params).json()
 
+    def get_hrv_data(self, cdate: str) -> Dict[str, Any]:
+        """Return Heart Rate Variability (hrv) data for current user."""
+
+        url = f"{self.garmin_connect_hrv_url}/{cdate}"
+        logger.debug("Requesting Heart Rate Variability (hrv) data")
+
+        return self.modern_rest_client.get(url).text #.json()
+
     def get_training_readiness(self, cdate: str) -> Dict[str, Any]:
         """Return training readiness data for current user."""
 
@@ -617,7 +634,9 @@ class Garmin:
         devices = self.get_devices()
         for device in devices:
             device_settings = self.get_device_settings(device["deviceId"])
-            alarms += device_settings["alarms"]
+            device_alarms = device_settings["alarms"]
+            if device_alarms is not None:
+                alarms += device_alarms
         return alarms
 
     def get_device_last_used(self):
@@ -646,12 +665,22 @@ class Garmin:
 
         return None
 
-    def upload_fit_activity(self, fit_file):
+    def upload_activity(self, activity_path: str):
         """Upload activity in fit format from file."""
+        # This code is borrowed from python-garminconnect-enhanced ;-)
+        file_base_name = os.path.basename(activity_path)
+        file_extension = file_base_name.split(".")[-1]
+        allowed_file_extension = file_extension.upper() in Garmin.ActivityUploadFormat.__members__
 
-        logger.debug("Uploading activity in fit format from file")
-        with open(fit_file, 'rb') as file:
-            return self.modern_rest_client.post(self.garmin_connect_fit_upload_url, {}, {}, file)
+        if allowed_file_extension:
+            files = {
+                "file": (file_base_name, open(activity_path, "rb" or "r")),
+            }
+
+            url = self.garmin_connect_upload
+            return self.modern_rest_client.post(url, files=files).json()
+        else:
+            raise GarminConnectInvalidFileFormatError(f"Could not upload {activity_path}")
 
     def get_activities_by_date(self, startdate, enddate, activitytype=None):
         """
@@ -692,8 +721,96 @@ class Garmin:
 
         return activities
 
+    def get_progress_summary_between_dates(self, startdate, enddate, metric="distance"):
+        """
+        Fetch progress summary data between specific dates
+        :param startdate: String in the format YYYY-MM-DD
+        :param enddate: String in the format YYYY-MM-DD
+        :param metric: metric to be calculated in the summary:
+            "elevationGain", "duration", "distance", "movingDuration"
+        :return: list of JSON activities with their aggregated progress summary
+        """
+
+        url = self.garmin_connect_fitnessstats
+        params = {
+            "startDate": str(startdate),
+            "endDate": str(enddate),
+            "aggregation": "lifetime",
+            "groupByParentActivityType": "true",
+            "metric": str(metric)
+            
+        }
+
+        logger.debug(
+            f"Requesting fitnessstats by date from {startdate} to {enddate}")
+        return self.modern_rest_client.get(url, params=params).json()
+
+    def get_activity_types(self):
+        url = self.garmin_connect_activity_types
+        logger.debug(f"Requesting activy types")
+        return self.modern_rest_client.get(url).json()
+
+    def get_goals(self, status="active", start=1, limit=30):
+        """
+        Fetch all goals based on status
+        :param status: Status of goals (valid options are "active", "future", or "past")
+        :type status: str
+        :param start: Initial goal index
+        :type start: int
+        :param limit: Pagination limit when retrieving goals
+        :type limit: int
+        :return: list of goals in JSON format
+        """
+
+        goals = []
+        url = self.garmin_connect_goals_url
+        params = {
+            "status": status,
+            "start": str(start),
+            "limit": str(limit),
+            "sortOrder": "asc"
+        }
+
+        logger.debug(f"Requesting {status} goals")
+        while True:
+            params["start"] = str(start)
+            logger.debug(f"Requesting {status} goals {start} to {start + limit - 1}")
+            goals_json = self.modern_rest_client.get(url, params=params).json()
+            if goals_json:
+                goals.extend(goals_json)
+                start = start + limit
+            else:
+                break
+
+        return goals
+
+    def get_gear(self, userProfileNumber):
+        """Return all user gear."""
+        url = f"{self.garmin_connect_gear}?userProfilePk={userProfileNumber}"
+        logger.debug("Requesting gear for user %s", userProfileNumber)
+
+        return self.modern_rest_client.get(url).json()
+
+    def get_gear_stats(self, gearUUID):
+        url = f"{self.garmin_connect_gear_baseurl}stats/{gearUUID}"
+        logger.debug("Requesting gear stats for gearUUID %s", gearUUID)
+        return self.modern_rest_client.get(url).json()
+
+    def get_gear_defaults(self, userProfileNumber):
+        url = f"{self.garmin_connect_gear_baseurl}user/{userProfileNumber}/activityTypes"
+        logger.debug("Requesting gear for user %s", userProfileNumber)
+        return self.modern_rest_client.get(url).json()
+
+    def set_gear_default(self, activityType, gearUUID, defaultGear=True):
+        defaultGearString = "/default/true" if defaultGear else ""
+        method_override = "PUT" if defaultGear else "DELETE"
+        url = f"{self.garmin_connect_gear_baseurl}{gearUUID}/activityType/{activityType}{defaultGearString}"
+        return self.modern_rest_client.post(
+            url, {"x-http-method-override": method_override}
+        )
+
     class ActivityDownloadFormat(Enum):
-        """Activitie variables."""
+        """Activity variables."""
 
         ORIGINAL = auto()
         TCX = auto()
@@ -701,6 +818,11 @@ class Garmin:
         KML = auto()
         CSV = auto()
 
+    class ActivityUploadFormat(Enum):
+        FIT = auto()
+        GPX = auto()
+        TCX = auto()
+
     def download_activity(self, activity_id, dl_fmt=ActivityDownloadFormat.TCX):
         """
         Downloads activity in requested format and returns the raw bytes. For
@@ -781,6 +903,15 @@ class Garmin:
 
         return self.modern_rest_client.get(url, params=params).json()
 
+    def get_activity_exercise_sets(self, activity_id):
+      """Return activity exercise sets."""
+
+      activity_id = str(activity_id)
+      url = f"{self.garmin_connect_activity}/{activity_id}/exerciseSets"
+      logger.debug("Requesting exercise sets for activity id %s", activity_id)
+
+      return self.modern_rest_client.get(url).json()
+
     def get_activity_gear(self, activity_id):
         """Return gears used for activity id."""
 
@@ -809,3 +940,7 @@ class GarminConnectTooManyRequestsError(Exception):
 
 class GarminConnectAuthenticationError(Exception):
     """Raised when authentication is failed."""
+
+
+class GarminConnectInvalidFileFormatError(Exception):
+    """Raised when an invalid file format is passed to upload."""
\ No newline at end of file
